/*
 * OsmCut.java
 *
 * Created on 10. August 2007, 09:37
 *
 * This tool is created to cut the OpenStreetMap Planetfile in various slices.
 * Please keep in mind, that the Planetfile has to be free of all xml errors.
 *
 * I know this is ugly code. If you are delicate, you should not look at this code
 * for your own safety.
 * 
 * The software is licensed under the GPL version 2. The maps it produces are to be 
 * licensed according to the data they were derived from which is 
 * Creative Commons CC-BY-SA 2.0 for OSM data. There is therefore no licensing 
 * conflict that might exist with other approaches.
 */

package osmcut;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.HashMap;

/**
 * @author Norman Rieß <norman@smash-net.org>
 * @thinker Carsten Schwede <computerteddy@gmx.de>
 */
public class OsmCut {
    

    public OsmCut() 
    {
    
    }
    
    /**
     * @param args the command line arguments (<size_in_°> <osm-file> <directory>)
     */
    public static void main(String[] args)
    {
        
        /***********************************************************************************************
         * Einlesen der Argumente
         *
         */
        
        float ksize = 3;
        String fileName = "";
        String resultDirectoryPath = "";
        
        if(args.length == 3)
        {
            ksize = Float.parseFloat(args[0]);
            fileName = args[1];
            resultDirectoryPath = args[2];
        }
        else
        {
            System.out.println("Bitte richtige Argumente übergeben. <Kachelgroesse in °> <Quelldatei> <Zielordner>");
            System.exit(1);
        }
        
        /***********************************************************************************************
         * Erzeugen des Ergebnisordners
         *
         */
        
        File resultDirectory = new File(resultDirectoryPath);
        if(!resultDirectory.exists())
        {
            resultDirectory.mkdir();
        }
        
        /***********************************************************************************************
         * HashMaps
         *  
         */
        
        // Nodes HashMap Key = ID, Wert = DateiNummer
        HashMap<Integer, Integer> nodes = new HashMap<Integer, Integer>();
        
        // Files HashMap Key = DateiNummer , Wert = PrintWriter
        HashMap<Integer, PrintWriter> files = new HashMap<Integer, PrintWriter>();
        
        // Segments HashMap Key = ID, Wert = DateiNummern
        HashMap<Integer, Integer> segments = new HashMap<Integer, Integer>();
        
        try 
        {
            
            /***********************************************************************************************
             * Variablen
             *  
             */
            
            BufferedReader in = new BufferedReader(new FileReader(fileName));
            
            String line;
            ArrayList xmlBox = new ArrayList();
            ArrayList wayFileNumbers = new ArrayList();
            String[] xmlHeader = new String[2];
            Integer[] segmFileNumber = new Integer[2];
            int id, fileNumber, from, to;
            double lat, lon;
            PrintWriter out;
            
            
            /***********************************************************************************************
             * Beginn Einlesen der Datei
             *  
             */
            
            xmlHeader[0] = in.readLine();
            xmlHeader[1] = in.readLine();
            while((line = in.readLine()) != null)
            {
                /***********************************************************************************
                 * Nodebehandlung
                 *
                 */

                if(line.trim().startsWith("<node"))
                {
                    if(line.trim().endsWith("/>"))
                    {
                        xmlBox.add(line);
                    }
                    else
                    {
                        xmlBox.add(line);
                        while(!((line = in.readLine()).trim()).startsWith("</node>"))
                        {
                            xmlBox.add(line);   
                        }
                        xmlBox.add(line); // </node> einlesen.
                    }
                    
                    id = Integer.parseInt(((String) xmlBox.get(0)).substring(((String) xmlBox.get(0)).indexOf("id=") + 4, ((String) xmlBox.get(0)).indexOf("\"", ((String) xmlBox.get(0)).indexOf("id=") + 4 )));
                    lat = Double.parseDouble(((String) xmlBox.get(0)).substring(((String) xmlBox.get(0)).indexOf("lat=") + 5, ((String) xmlBox.get(0)).indexOf("\"", ((String) xmlBox.get(0)).indexOf("lat=") + 5 )));
                    lon = Double.parseDouble(((String) xmlBox.get(0)).substring(((String) xmlBox.get(0)).indexOf("lon=") + 5, ((String) xmlBox.get(0)).indexOf("\"", ((String) xmlBox.get(0)).indexOf("lon=") + 5 )));
                    
                    //fileNumber = (((int)(lat + 90) / 3) + ((int)(lon + 180) / 3) * 60) + 1;
                    fileNumber = ((int)((lat + 90) / ksize) + (int)((lon + 180) / ksize) * (int)(180 / ksize)) + 1;

                    nodes.put(id, fileNumber);
                    
                    if((out = files.get(fileNumber)) == null)
                    {
                        out = new PrintWriter(new FileWriter(resultDirectoryPath + "/" + (63240000 + fileNumber), true));
                        files.put(fileNumber, out);
                        out.println(xmlHeader[0]);
                        out.println(xmlHeader[1]);
                    }
                                        
                    for(int i = 0; i < xmlBox.size(); i++)
                    {
                        out.println((String)xmlBox.get(i));
                    }
                    out.flush();
                    
                }
                
                /***********************************************************************************
                 * Segmentbehandlung
                 *
                 */
                
                else if(line.trim().startsWith("<segment"))
                {
                   
                   if(line.trim().endsWith("/>"))
                    {
                        xmlBox.add(line);
                    }
                    else
                    {
                        xmlBox.add(line);
                        while(!((line = in.readLine()).trim()).startsWith("</segment>"))
                        {
                            xmlBox.add(line);   
                        }
                        xmlBox.add(line); // </segment> einlesen.
                    }
                   
                   id = Integer.parseInt(((String) xmlBox.get(0)).substring(((String) xmlBox.get(0)).indexOf("id=") + 4, ((String) xmlBox.get(0)).indexOf("\"", ((String) xmlBox.get(0)).indexOf("id=") + 4 )));
                   from = Integer.parseInt(((String) xmlBox.get(0)).substring(((String) xmlBox.get(0)).indexOf("from=") + 6, ((String) xmlBox.get(0)).indexOf("\"", ((String) xmlBox.get(0)).indexOf("from=") + 6 )));
                   to = Integer.parseInt(((String) xmlBox.get(0)).substring(((String) xmlBox.get(0)).indexOf("to=") + 4, ((String) xmlBox.get(0)).indexOf("\"", ((String) xmlBox.get(0)).indexOf("to=") + 4 )));
                    
                   segmFileNumber[0] = nodes.get(from);
                   segmFileNumber[1] = nodes.get(to);
                   
                   if(segmFileNumber[0].compareTo(segmFileNumber[1]) == 0)
                   {
                        segments.put(id, segmFileNumber[0]);
                        out = files.get(segmFileNumber[0]);
                        for(int i = 0; i < xmlBox.size(); i++)
                        {
                            out.println((String)xmlBox.get(i));
                        }
                        out.flush();
                   }
   
                }
                
                /***********************************************************************************
                 * Waybehandlung
                 *
                 */
                
                else if(line.trim().startsWith("<way"))
                {
                    xmlBox.add(line);
                    while(!((line = in.readLine()).trim()).startsWith("</way>"))
                    {
                        xmlBox.add(line);   
                    }
                    xmlBox.add(line); // </way> einlesen.

                    for(int i = 0; i < xmlBox.size(); i++)
                    {
                        if(((String) xmlBox.get(i)).trim().startsWith("<seg"))
                        {
                            id = Integer.parseInt(((String) xmlBox.get(i)).substring(((String) xmlBox.get(i)).indexOf("id=") + 4, ((String) xmlBox.get(i)).indexOf("\"", ((String) xmlBox.get(i)).indexOf("id=") + 4 )));
                            segmFileNumber[0] = segments.get(id);

                            if(segmFileNumber[0] != null)
                            {
                                if(!wayFileNumbers.contains(segmFileNumber[0]))
                                {
                                    wayFileNumbers.add(segmFileNumber[0]);
                                }
                            }
                        }                            
                    }

                    for(int i = 0; i < wayFileNumbers.size(); i++)
                    {
                        out = files.get((Integer)wayFileNumbers.get(i));
                        for(int l = 0; l < xmlBox.size(); l++)
                        {
                            out.println((String)xmlBox.get(l));
                        }
                        out.flush();
                    }
                }
                
                /***********************************************************************************
                 * Löschen der ArrayListen am Ende der Schleife
                 *
                 */
                
                wayFileNumbers.clear();
                xmlBox.clear();
                
                //***********************************************************************************
            }
            
            
            /***********************************************************************************
             * Abschließen und Schließen der Reader und FileWriter
             *
             */
            in.close();

            // Alle Dateien abschließen und schließen.
            for(Object fe : files.values().toArray())
            {
                ((PrintWriter)fe).println("</osm>");
                ((PrintWriter)fe).flush();
                ((PrintWriter)fe).close();
            }
            //**********************************************************************************
        }
        catch (FileNotFoundException ex) 
        {
            System.out.println("Datei nicht gefunden. " + fileName);
            System.exit(1);
        }
        catch (IOException ex) 
        {
            ex.printStackTrace();
        }
         
    }
    
}
